################################################################################
#
# Point process modeling via process convolutions
# Copyright (C) 2011, University of California
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Questions? Contact Waley W. Liang (wliang@ams.ucsc.edu)
#
################################################################################

library("mvtnorm", verbose = FALSE)

## Shortcuts for build.kernel
build.kernel.bezier <-
  function(xy, mu.xy, l, w)
{
  return(build.kernel(xy, mu.xy, k.bezier, l, w))
}


build.kernel.normal <- 
  function(xy, mu.xy, sigma)
{
  return(build.kernel(xy, mu.xy, k.normal, sigma = sigma))
}


## Builds an arbitrary kernel matrix
build.kernel <- 
  function(xy, mu.xy, kernel, ...)
{
  x <- xy[, 1]
  y <- xy[, 2]
  mu.x <- mu.xy[, 1]
  mu.y <- mu.xy[, 2]
  nrow <- length(x)
  ncol <- length(mu.x)
  
  if(nrow != length(x))    warning("nrow != length(x)")
  if(nrow != length(y))    warning("nrow != length(y)")
  if(ncol != length(mu.x)) warning("ncol != length(mu.x)")
  if(ncol != length(mu.y)) warning("ncol != length(mu.y)")
  return(matrix(kernel(cbind(as.vector(outer(x, mu.x, "-")), 
                             as.vector(outer(y, mu.y, "-"))), ...),
                             nrow = nrow, ncol = ncol))
}


k.normal <- 
  function(x, sigma)
{
  return(dmvnorm(x, c(0, 0), sigma))
}


k.bezier <- 
  function(x, l = c(1, 1), w = 3) 
{
  k.star2 <- function(x, l = c(1, 1), w = 3){
    max(1 - sum((x / l)^2), 0)^w
  }
  k.coef2 <- 1
  if(is.vector(x)) x <- matrix(x, nrow = 1)
  return(k.coef2 * apply(x, 1, k.star2, l, w))
}


## Builds the time component F'(t)
time.component <- 
  function(weekday, month, fit.weekday, fit.month, phi.init, burnin, iter,
           num.months)
{
  if(fit.weekday & !fit.month){
    phi <- matrix(0, nrow = 2, ncol = iter + burnin)
    phi[, 1] <- phi.init[1:2]
    return(list(Ft.cal = cbind(sin(2 * pi * seq(1, 7, 1) / 7), 
                               cos(2 * pi * seq(1, 7, 1) / 7)), 
                Ft.obsv = c(sum(sin(2 * pi * weekday / 7)), 
                            sum(cos(2 * pi * weekday / 7))), 
                phi = phi))
  }
  else if(!fit.weekday & fit.month){
    phi <- matrix(0, nrow = 2, ncol = iter + burnin)
    phi[, 1] <- phi.init[3:4]
    return(list(Ft.cal = cbind(sin(2 * pi * seq(1, num.months, 1) / 12), 
                               cos(2 * pi * seq(1, num.months, 1) / 12)), 
                Ft.obsv = c(sum(sin(2 * pi * month / 12)), 
                            sum(cos(2 * pi * month / 12))), 
                phi = phi))
  }
  else if(fit.weekday & fit.month){
    phi <- matrix(0, nrow = 4, ncol = iter + burnin)
    phi[, 1] <- phi.init
    cal <- expand.grid(seq(1, 7, 1), seq(1, num.months, 1))
    return(list(Ft.cal = cbind(sin(2 * pi * cal[, 1] / 7), 
                               cos(2 * pi * cal[, 1] / 7), 
                               sin(2 * pi * cal[, 2] / 12), 
                               cos(2 * pi * cal[, 2] / 12)), 
                Ft.obsv = c(sum(sin(2 * pi * weekday / 7)), 
                            sum(cos(2 * pi * weekday / 7)), 
                            sum(sin(2 * pi * month / 12)), 
                            sum(cos(2 * pi * month / 12))), 
                phi = phi))
  }
  else
    return(NULL)
}


## Update the Lambda.0 term in the static model
update.Lambda.0 <- 
  function(K.grid, zu, us.col, grid.area)
{
  return(as.vector(t(K.grid %*% zu) %*% (us.col * grid.area)))
}


## Update the Lambda.t term in the static model
update.Lambda.T <- 
  function(vt.col)
{
  return(sum(vt.col))
}


to.utm <- function(x.ll) {
  x.ll <- cbind(1:nrow(x.ll), x.ll)
  dimnames(x.ll) <- list(NULL, c("PID", "X", "Y"))
  x.ll <- as.PolyData(x.ll, projection = "LL", zone = 14)
  S <- convUL(x.ll)
  S <- as.matrix(S[, c(2, 3)])
  return(S)
}


## cov.inv is the precision matrix
mahalanobis.ppm <-
  function(x, center, cov.inv)
{
  x <- if (is.vector(x))
    matrix(x, ncol = length(x))
  else as.matrix(x)
  x <- sweep(x, 2, center)
  retval <- rowSums((x %*% cov.inv) * x)
  names(retval) <- rownames(x)
  retval
}


## Modified from package mvtnorm
dmvnormal <-
  function(locs, mean, cov.inv, log = FALSE)
{
  if (is.vector(locs)) {
    locs <- matrix(locs, ncol = length(locs))
  }
  if (missing(mean)) {
    mean <- rep(0, length = ncol(locs))
  }
  if (missing(cov.inv)) {
    cov.inv <- diag(ncol(locs))
  }
  if (NCOL(locs) != NCOL(cov.inv)) {
    stop("locs and cov.inv have non-conforming size")
  }
  if (!isSymmetric(cov.inv, tol = sqrt(.Machine$double.eps))) {
    stop("cov.inv must be a symmetric matrix")
  }
  if (length(mean) != NROW(cov.inv)) {
    stop("mean and cov.inv have non-conforming size")
  }
  distval <- mahalanobis.ppm(x = locs, center = mean,
                               cov.inv = cov.inv)
  logdet <- sum(log(eigen(cov.inv, symmetric = TRUE,
                          only.values = TRUE)$values))
  logretval <- -(ncol(locs) * log(2 * pi) - logdet + distval) / 2
  if (log)
    return(logretval)
  exp(logretval)
}


## Bezier kernel
dmvbezier <-
  function(locs, center, cov.inv, kw)
{
  s.norm <- mahalanobis.ppm(x = locs, center = center,
                            cov.inv = cov.inv)
  s.norm[s.norm > 1] <- 1
  return((1 - s.norm)^kw)
}


sample.beta <-
  function(beta.old, pri.beta.mean, pri.beta.sd, K.grid.s, zu, us.col, vt.col,
           grid.area, Lambda.0, Lambda.T, Xs.obsv, tau, pro.beta.sd, Xs.grid,
           num.beta, num.grid, Lambda.dlm, num.months, dlm, K.zu.grid) 
{
  accept <- rep(0, num.beta)
  beta.new <- beta.old
  for (j in 1:num.beta) {
    beta.new[j] <- rnorm(1, beta.old[j], pro.beta.sd[j])
    us.col.new <- exp(Xs.grid %*% beta.new)
    log.pri.diff <- -((beta.new[j] - pri.beta.mean[j])^2 -
                      (beta.old[j] - pri.beta.mean[j])^2) / (2 * pri.beta.sd[j]^2)
    if (dlm) {
      Lambda.dlm.new <- sum(colSums(K.zu.grid * as.vector(us.col.new * grid.area)) * vt.col)
      log.lik.diff <- -tau * (Lambda.dlm.new - Lambda.dlm) +
                       (beta.new[j] - beta.old[j]) * Xs.obsv[j]
    } else {
      Lambda.0.new <- update.Lambda.0(K.grid.s, zu, us.col.new, grid.area)      
      log.lik.diff <- -tau * Lambda.T * (Lambda.0.new - Lambda.0) +
                       (beta.new[j] - beta.old[j]) * Xs.obsv[j]
    }
    a <- min(1, exp(log.lik.diff + log.pri.diff))
    
    if (a > runif(1)) {
      us.col <- us.col.new
      accept[j] <- accept[j] + 1
      if (dlm)
        Lambda.dlm <- Lambda.dlm.new
      else
        Lambda.0 <- Lambda.0.new
    } else {
      beta.new[j] <- beta.old[j]
    }
  }
  return(list(Lambda.0 = Lambda.0, Lambda.dlm = Lambda.dlm, beta = beta.new, 
              us.col = us.col, accept = accept))
}


## Perform Metropolis Hastings algorithm for phi
sample.phi <-
  function(phi.old, pri.phi.sd, us.col, vt.col, zu, tau, K.grid.s, Ft.obsv,
           Ft.cal, Lambda.0, Lambda.T, Lambda.dlm, pro.phi.sd, num.phi,
           num.grid, grid.area, num.months, dlm, K.zu.grid) 
{
  accept <- rep(0, num.phi)
  phi.new <- phi.old
  for (j in 1:num.phi) {
    phi.new[j] <- rnorm(1, phi.old[j], pro.phi.sd[j])
    vt.col.new <- exp(Ft.cal %*% phi.new)
    log.pri.diff <- -(phi.new[j]^2 - phi.old[j]^2) / (2 * pri.phi.sd[j]^2)
    if (dlm) {
      Lambda.dlm.new <- sum(colSums(K.zu.grid * as.vector(us.col * grid.area)) * vt.col.new)
      log.lik.diff <- -tau * (Lambda.dlm.new - Lambda.dlm) + (phi.new[j] - phi.old[j]) * Ft.obsv[j]
    } else {
      Lambda.T.new <- update.Lambda.T(vt.col.new)
      log.lik.diff <- -tau * (Lambda.T.new - Lambda.T) * Lambda.0 +
                       (phi.new[j] - phi.old[j]) * Ft.obsv[j]
    }
    a <- min(1, exp(log.lik.diff + log.pri.diff))
    if (a > runif(1)) { 
      vt.col <- vt.col.new
      accept[j] <- accept[j] + 1
      if (dlm)
        Lambda.dlm <- Lambda.dlm.new
      else
        Lambda.T <- Lambda.T.new
    } else {
      phi.new[j] <- phi.old[j]
    }
  }
  return(list(Lambda.T = Lambda.T, Lambda.dlm = Lambda.dlm, phi = phi.new,
              vt.col = vt.col, accept = accept))
}


## Performs Metropolis Hastings algorithm for zu's
sample.zu <- 
  function(zu.old, pro.zu.a, pro.zu.del, Lambda.0, pri.zu.shape, pri.zu.rate, 
           K.obsv.m, K.obsv.s, tau, Lambda.T, us.col, vt.col, grid.area,
           num.kern, K.grid.m, K.grid.s, proposal = 3, theta.bs, pri.zu.sd,
           Lambda.dlm, num.months, dlm) 
{
  accept <- rep(0, length(zu.old))
  ## Proposal 1: log-normal
  if (proposal == 1) {
    sd.lnorm <- 0.25
    zu.new <- rlnorm(length(zu.old), meanlog = log(zu.old), sdlog = sd.lnorm)
    index.zu <- rep(FALSE, length(zu.old))
  }
  ## Proposal 2: beta
  else if (proposal == 2) {
    pro.alpha <- pro.zu.a * pro.zu.del / 2
    pro.beta <- pro.zu.a * (1 - pro.zu.del) / 2
    v <- rbeta(length(zu.old), pro.alpha, pro.beta)
    zu.new <- zu.old * v / pro.zu.del
    
    ## Ensure reverse jump probablity is in range
    index.zu <- ((pro.zu.del * zu.old / zu.new) >= 1) | ((pro.zu.del * zu.old / zu.new) <= 0)
    zu.new[index.zu] <- zu.old[index.zu]
  }
  else
    warning("Bad proposal")
  
  ## Initialize K.zu.obsv.old
  if (dlm) {
    K.zu.grid <- matrix(0, nrow = nrow(K.grid.m), ncol = num.months)
    K.zu.obsv.old <- NULL
    for (t in 1:num.months) {
      ind <- ((t - 1) * num.kern + 1):(t * num.kern)
      K.zu.obsv.old <- c(K.zu.obsv.old, list(K.obsv.s[[t]] %*% zu.old[ind]))
    }
  } else {
    K.zu.grid <- NULL
    K.zu.obsv.old <- K.obsv.s %*% zu.old
  }
  
  ## Start sampling
  if (dlm) {
    for (t in 1:num.months) {
      for (i in 1:num.kern) {
        ind <- (t - 1) * num.kern + i
        if (!(index.zu[ind])) {
          K.zu.obsv.new.t <- K.zu.obsv.old[[t]] + K.obsv.m[[t]][, i] * (zu.new[ind] - zu.old[ind])
          Lambda.dlm.diff <- as.numeric(crossprod(K.grid.m[, i], (us.col * grid.area)) *
                                        (zu.new[ind] - zu.old[ind]) * vt.col[t])
          
          log.lik.diff <- -tau * Lambda.dlm.diff + (sum(log(K.zu.obsv.new.t) - log(K.zu.obsv.old[[t]])))
          log.pri.diff <- (-log(zu.new[ind]) + dnorm(log(zu.new[ind]), theta.bs[ind], pri.zu.sd, log = TRUE)) -
                          (-log(zu.old[ind]) + dnorm(log(zu.old[ind]), theta.bs[ind], pri.zu.sd, log = TRUE))
          
          if (proposal == 1) {
            log.pro.diff <- dlnorm(zu.old[i], log(zu.new[i]), sd.lnorm, log = TRUE) -
                            dlnorm(zu.new[i], log(zu.old[i]), sd.lnorm, log = TRUE)
          } else if (proposal == 2) { 
            log.pro.diff <- log(zu.old[ind] / zu.new[ind]) + 
              dbeta(pro.zu.del * zu.old[ind] / zu.new[ind], pro.alpha, pro.beta, log = TRUE) -
              dbeta(pro.zu.del * zu.new[ind] / zu.old[ind], pro.alpha, pro.beta, log = TRUE)
          }
          alpha <- min(1, exp(log.lik.diff + log.pri.diff + log.pro.diff))
          if (is.nan(alpha)) {
            zu.new[ind] <- zu.old[ind]
          }
          else if (runif(1) > alpha) { ## if rejected
            zu.new[ind] <- zu.old[ind]
          }
          else {
            K.zu.obsv.old[[t]] <- K.zu.obsv.new.t
            accept[ind] <- accept[ind] + 1
            Lambda.dlm <- Lambda.dlm + Lambda.dlm.diff
          }
        }
      }
      ind <- ((t - 1) * num.kern + 1):(t * num.kern)
      K.zu.grid[, t] <- K.grid.m %*% zu.new[ind]
    }
  } else {
    for (i in 1:num.kern) {
      if (!(index.zu[i])) {
        K.zu.obsv.new <- K.zu.obsv.old + K.obsv.m[, i] * (zu.new[i] - zu.old[i])
        Lambda.0.diff <- as.numeric(crossprod(K.grid.m[, i], (us.col * grid.area)) *
                                    (zu.new[i] - zu.old[i]))
        log.lik.diff <- -tau * Lambda.T * Lambda.0.diff +
                        (sum(log(K.zu.obsv.new) - log(K.zu.obsv.old)))
        log.pri.diff <- (pri.zu.shape - 1) * (log(zu.new[i]) - log(zu.old[i])) -
                        pri.zu.rate * (zu.new[i] - zu.old[i])
        if (proposal == 1) {
          log.pro.diff <- dlnorm(zu.old[i], log(zu.new[i]), sd.lnorm, log = TRUE) -
                          dlnorm(zu.new[i], log(zu.old[i]), sd.lnorm, log = TRUE)
        } else if (proposal == 2) { 
          log.pro.diff <- log(zu.old[i] / zu.new[i]) + 
            dbeta(pro.zu.del * zu.old[i] / zu.new[i], pro.alpha, pro.beta, log = TRUE) -
            dbeta(pro.zu.del * zu.new[i] / zu.old[i], pro.alpha, pro.beta, log = TRUE)
        }
        alpha <- min(1, exp(log.lik.diff + log.pri.diff + log.pro.diff))
        if (is.nan(alpha)) {
          zu.new[i] <- zu.old[i]
        }
        else if (runif(1) > alpha) { ## if rejected
          zu.new[i] <- zu.old[i]
        }
        else {
          K.zu.obsv.old <- K.zu.obsv.new
          accept[i] <- accept[i] + 1
          Lambda.0 <- Lambda.0 + Lambda.0.diff
        }
      }
    }
  }
  return(list(Lambda.0 = Lambda.0, Lambda.dlm = Lambda.dlm, zu = zu.new,
              accept = accept, K.zu.grid = K.zu.grid))
}


main <-
  function(domain = NULL, grid.data.file = NULL, crime.data.file = NULL, 
           output.file = NULL, epsilon = .Machine$double.eps,
           burnin = 1000, iter = 2000, sigma = diag(c(0.007, 0.0035)^2), 
           num.kern.lon = 20, num.kern.lat = 10, proposal = 1,
           kernel.fun = "bezier", kernel.l = 3, kernel.w = 3,           
           popden.scale = 1 / 30000, vacant.scale = 1, sparse = TRUE, fit.weekday = FALSE,
           fit.month = TRUE, pri.zu.shape = 1, pri.zu.rate = 1, pri.beta.mean = c(0, 0),
           pri.beta.sd = c(1, 1), pri.phi.sd = c(1, 1), pro.zu.a = 15,
           pro.zu.del = 0.5, pro.beta.sd = c(0.15, 0.25), pro.phi.sd = c(0.1, 0.1), 
           tau = 1, zu.init = 0.001, beta.init = rep(0, 0), phi.init = rep(0, 0),
           dlm = FALSE, v = 0.1, m0 = log(0.001), c0 = 1, alpha0 = 1, beta0 = 0.001,
           del.star = 0.95, discount = 0.95, predict.month = NULL)
{
  ## Only Bezier kernel is supported.
  if (kernel.fun != "bezier") {
    kernel.fun <- "bezier"
    cat("Warnings: Only Bezier kernel is supported.\n")
  }
  
  data <- pre.process(grid.data.file = grid.data.file, crime.data.file = crime.data.file, 
                      num.kern.lon = num.kern.lon, num.kern.lat = num.kern.lat,
                      popden.scale = popden.scale, vacant.scale = vacant.scale, domain = domain)
  corv <- data$corv
  obsv <- data$obsv
  kernel <- data$kernel
  num.grid <- nrow(corv)
  num.obsv <- nrow(obsv)
  num.kern <- nrow(kernel)
  num.beta <- length(beta.init)
  num.phi <- 4
  obsv.list <- NULL
  num.obsv.list <- NULL
  K.grid.list <- NULL
  K.obsv.list <- NULL
  num.months <- 12
  obsv.month <- obsv$month
  obsv.weekday <- obsv$weekday
  if (dlm) {
    fit.weekday <- FALSE
    fit.month <- TRUE
    for (t in 1:num.months) {
      obsv.list <- c(obsv.list, list(obsv[obsv$month == t, ]))
      num.obsv.list <- c(num.obsv.list, nrow(obsv.list[[t]]))
    }
    if (!is.null(predict.month)) {
      num.months <- (predict.month - 1)
      obsv.month <- obsv$month[obsv$month < predict.month]
      obsv.weekday <- obsv$weekday[obsv$month < predict.month]
    }
  }
  
  ## Build kernel matrices:
  if (kernel.fun == "normal") {
    K.grid.m <- build.kernel.normal(corv[c("lon", "lat")], kernel, sigma)
    K.grid.m <- K.grid.m / apply(K.grid.m, 1, sum)
    if (epsilon > 0) {
      K.grid.m[K.grid.m < epsilon] <- 0
    }
    if (dlm) {
      for (t in 1:num.months) {
        K.obsv.m <- build.kernel.normal(obsv.list[[t]], kernel, sigma)
        K.obsv.m <- K.obsv.m / apply(K.obsv.m, 1, sum)
        if (epsilon > 0) {
          K.obsv.m[K.obsv.m < epsilon] <- 0
        }      
        K.obsv.list <- c(K.obsv.list, list(K.obsv.m))
      }
      K.obsv.m <- K.obsv.list
      rm(K.obsv.list)
      gc()
    }
    else {
      K.obsv.m <- build.kernel.normal(obsv, kernel, sigma)
      K.obsv.m <- K.obsv.m / apply(K.obsv.m, 1, sum)
      if (epsilon > 0) {
        K.obsv.m[K.obsv.m < epsilon] <- 0
      }
    }
  } else if (kernel.fun == "bezier") {
    K.grid.m <- apply(as.matrix(kernel), 1, function(center, locs)
                      dmvbezier(locs, center, diag(1 / (kernel.l * data$delta)^2), kernel.w),
                      as.matrix(corv[c("lon", "lat")]))
    K.grid.m <- K.grid.m / apply(K.grid.m, 1, sum)
    if (epsilon > 0) {
      K.grid.m[K.grid.m < epsilon] <- 0
    }
    if (dlm) {
      for (t in 1:num.months) {
        K.obsv.m <- apply(as.matrix(kernel), 1, function(center, locs)
                          dmvbezier(locs, center, diag(1 / (kernel.l * data$delta)^2), kernel.w),
                          as.matrix(obsv.list[[t]][c("lon", "lat")]))
        K.obsv.m <- K.obsv.m / apply(K.obsv.m, 1, sum)
        if (epsilon > 0) {
          K.obsv.m[K.obsv.m < epsilon] <- 0
        }
        K.obsv.list <- c(K.obsv.list, list(K.obsv.m))
      }
      K.obsv.m <- K.obsv.list
      rm(K.obsv.list)
      gc()
    } else {
      K.obsv.m <- apply(as.matrix(kernel), 1, function(center, locs)
                        dmvbezier(locs, center, diag(1 / (kernel.l * data$delta)^2), kernel.w),
                        as.matrix(obsv[c("lon", "lat")]))
      K.obsv.m <- K.obsv.m / apply(K.obsv.m, 1, sum)
      if (epsilon > 0) {
        K.obsv.m[K.obsv.m < epsilon] <- 0
      }
    }
  }
  else
    stop("Bad kernel.fun")

  K.zu.grid <- matrix(0, nrow = num.grid, ncol = num.months)
  if (dlm) {
    for (t in 1:num.months) {
      K.zu.grid[, t] <- K.grid.m %*% rep(zu.init, num.kern) 
    }
  }
  K.obsv.s <- K.obsv.m
  K.grid.s <- K.grid.m
  
  if (num.beta == 2) {
    Xs.grid <- cbind(corv$popden, corv$vacant)
    Xs.obsv <- c(sum(obsv$popden), sum(obsv$vacant))
  }
  else {
    Xs.grid <- cbind(rep(1, num.grid), corv$popden, corv$vacant)
    Xs.obsv <- c(num.obsv, sum(obsv$popden), sum(obsv$vacant))
  }  
  
  Lambda.dlm <- NULL
  Lambda.0 <- NULL
  Lambda.T <- 1
  Lambda <- matrix(nrow = burnin + iter, ncol = 2)
  
  if (dlm) {
    zu <- matrix(zu.init, nrow = num.kern*num.months, ncol = iter + burnin)
    zu.pred <- matrix(0.0, nrow = num.kern, ncol = iter + burnin)
  }
  else {
    zu <- matrix(zu.init, nrow = num.kern, ncol = iter + burnin)
    zu.pred <- NULL
  }
  theta.ff <- log(zu)
  theta.bs <- theta.ff  
  beta <- matrix(0, nrow = num.beta, ncol = iter + burnin)
  beta[, 1] <- beta.init
  us.col <- exp(Xs.grid %*% beta[, 1])
  
  phi <- NULL
  time <- time.component(obsv.weekday, obsv.month, fit.weekday, fit.month,
                         phi.init, burnin, iter, num.months)
  if (!is.null(time)) {
    Ft.cal <- time$Ft.cal
    Ft.obsv <- time$Ft.obsv
    phi <- time$phi
    num.phi <- nrow(phi)
    vt.col <- exp(Ft.cal %*% phi[, 1])
    if (!dlm) Lambda.T <- update.Lambda.T(vt.col)
  }
  if (dlm) {
    Lambda.dlm <- sum(colSums(K.zu.grid * as.vector(us.col * corv$grid.area)) * vt.col)
  }
  else
    Lambda.0 <- update.Lambda.0(K.grid.s, zu[, 1], us.col, corv$grid.area)
  
  accept.zu <- rep(0, nrow(zu))
  accept.beta <- rep(0, num.beta)
  accept.phi <- rep(0, num.phi)
  m0 <- rep(m0, num.kern)
  C0 <- rep(c0^2, num.kern)
  
  ## Start MCMC
  pb <- txtProgressBar(min = 1, max = (burnin + iter), style = 3)
  for (i in 2:(burnin + iter)) {
    if ((i %% as.integer((burnin + iter) / 100)) == 0)
      setTxtProgressBar(pb, i)
    
    zu.out <- sample.zu(zu[, i - 1], pro.zu.a, pro.zu.del, Lambda.0, pri.zu.shape,
                        pri.zu.rate, K.obsv.m, K.obsv.s, tau, Lambda.T, us.col,
                        vt.col, as.vector(corv$grid.area), num.kern, K.grid.m, K.grid.s,
                        proposal = proposal, theta.bs[, i - 1], v, Lambda.dlm,
                        num.months, dlm)
    zu[, i] <- zu.out$zu
    accept.zu <- accept.zu + zu.out$accept
    K.zu.grid <- zu.out$K.zu.grid

    if (dlm) {
      Lambda.dlm <- zu.out$Lambda.dlm
      ffbs.out <- ppmffbs(y = log(zu.out$zu), m0 = m0, C0 = C0, alpha0 = alpha0, beta0 = beta0,
                          T = num.months, n = num.kern, del.star = del.star, discount = discount)
      theta.ff[, i] <- as.vector(ffbs.out$m.ff)
      theta.bs[, i] <- as.vector(ffbs.out$m.bs)
      m0 <- ffbs.out$m.t
      C0 <- ffbs.out$C.t
      alpha0 <- ffbs.out$alpha.t
      beta0 <- ffbs.out$beta.t
      v <- sqrt(ffbs.out$v2)
      if (!is.null(predict.month))
        zu.pred[, i] <- exp(as.vector(ffbs.out$f))      
    }
    else
      Lambda.0 <- zu.out$Lambda.0
    
    beta.out <- sample.beta(beta[, i - 1], pri.beta.mean, pri.beta.sd, K.grid.s, zu[, i],
                            us.col, vt.col, as.vector(corv$grid.area), Lambda.0, Lambda.T,
                            Xs.obsv, tau, pro.beta.sd, Xs.grid, num.beta, num.grid,
                            Lambda.dlm, num.months, dlm, K.zu.grid)
    beta[, i] <- beta.out$beta
    us.col <- beta.out$us.col
    accept.beta <- accept.beta + beta.out$accept    
    
    if (dlm)
      Lambda.dlm <- beta.out$Lambda.dlm
    else
      Lambda.0 <- beta.out$Lambda.0
    
    if (!is.null(time)) {
      phi.out <- sample.phi(phi[, i - 1], pri.phi.sd, us.col, vt.col, zu[, i], tau,
                            K.grid.s, Ft.obsv, Ft.cal, Lambda.0, Lambda.T, Lambda.dlm,
                            pro.phi.sd, num.phi, num.grid, as.vector(corv$grid.area),
                            num.months, dlm, K.zu.grid)
      phi[, i] <- phi.out$phi
      vt.col <- phi.out$vt.col
      accept.phi <- accept.phi + phi.out$accept
      if (dlm)
        Lambda.dlm <- phi.out$Lambda.dlm
      else
        Lambda.T <- phi.out$Lambda.T
    }
    Lambda[i, ] <- c(Lambda.0, Lambda.T)
  }
  close(pb)

  zu <- zu[, (burnin + 1):(burnin + iter)]
  zu.mean <- apply(zu, 1, mean)
  zu.pred.mean <- NULL
  theta.ff <- theta.ff[, (burnin + 1):(burnin + iter)]
  theta.ff.mean <- apply(theta.ff, 1, mean)
  theta.bs <- theta.bs[, (burnin + 1):(burnin + iter)]
  theta.bs.mean <- apply(theta.bs, 1, mean)    
  beta <- beta[, (burnin + 1):(burnin + iter)]
  beta.mean <- apply(beta, 1, mean)
  phi.mean <- NULL
  if (!is.null(time)) {
    phi <- phi[, (burnin + 1):(burnin + iter)]
    phi.mean <- apply(phi, 1, mean)
  }
  baseline.fit <- NULL
  baseline.pred <- NULL
  baseline.forecast <- NULL
  spat.eff.fit <- NULL
  spat.eff.pred <- exp(Xs.grid %*% beta)
  spatial.cov <- as.matrix(data$obsv[c("popden", "vacant")])
  time.eff <- exp(Ft.cal %*% phi)

  if (dlm) {    
    for (t in 1:num.months) {
      start <- (t - 1) * num.kern + 1
      end <- t * num.kern      
      baseline.fit <- c(baseline.fit, list(tau * K.obsv.s[[t]] %*% zu[start:end, ]))
      baseline.pred <- c(baseline.pred, list(tau * K.grid.s %*% zu[start:end, ]))
      ind <- which(data$obsv$month == t)
      spat.eff.fit <- c(spat.eff.fit, list(exp(spatial.cov[ind, ] %*% beta)))
    }
    if (!is.null(predict.month)) {
      zu.pred <- zu.pred[, (burnin + 1):(burnin + iter)]
      zu.pred.mean <- apply(zu.pred, 1, mean)
      baseline.forecast <- tau * K.grid.s %*% zu.pred
      time.eff <- rbind(time.eff, exp(phi[1, ] * sin(2 * pi * predict.month / 12) +
                                      phi[2, ] * cos(2 * pi * predict.month / 12)))
    }
  } else {
    baseline.fit <- tau * K.obsv.s %*% zu
    baseline.pred <- tau * K.grid.s %*% zu
    spat.eff.fit <- exp(spatial.cov %*% beta)
  }
  
  mcmc <- list(zu = zu, zu.pred = zu.pred, tau = tau, beta = beta, phi = phi,
               zu.mean = zu.mean, zu.pred.mean = zu.pred.mean,
               beta.mean = beta.mean, phi.mean = phi.mean,
               accept.zu = accept.zu / (burnin + iter),
               accept.beta = accept.beta / (burnin + iter), 
               accept.phi = accept.phi / (burnin + iter),
               burnin = burnin, iter = iter,
               num.kern = num.kern, num.beta = num.beta,
               num.phi = num.phi, num.months = num.months,
               K.grid.m = K.grid.m, K.grid.s = K.grid.s,
               Xs.grid = Xs.grid, Xs.obsv = Xs.obsv,
               K.obsv.m = K.obsv.m, K.obsv.s = K.obsv.s,
               data = data, sigma = sigma, fit.weekday = fit.weekday,
               fit.month = fit.month, time = time,
               popden.scale = popden.scale, vacant.scale = vacant.scale,
               Lambda = Lambda, Lambda.dlm = Lambda.dlm,
               theta.ff = theta.ff, theta.ff.mean = theta.ff.mean,
               theta.bs = theta.bs, theta.bs.mean = theta.bs.mean,
               v = v, m0 = m0, c0 = c0, alpha0 = alpha0, beta0 = beta0,
               del.star = del.star, discount = discount, dlm = dlm,

               baseline.fit = baseline.fit, baseline.pred = baseline.pred,
               baseline.forecast = baseline.forecast,
               spat.eff.fit = spat.eff.fit, spat.eff.pred = spat.eff.pred,
               time.eff = time.eff,
               predict.month = predict.month, pri.zu.shape = pri.zu.shape,
               pri.zu.rate = pri.zu.rate, pri.beta.mean = pri.beta.mean,
               pri.beta.sd = pri.beta.sd, pri.phi.sd = pri.phi.sd,
               pro.zu.a = pro.zu.a, pro.zu.del = pro.zu.del,
               pro.beta.sd = pro.beta.sd, pro.phi.sd = pro.phi.sd,
               kernel.l = kernel.l, kernel.w = kernel.w)
  
  if (!is.null(output.file))
    save(mcmc, file = output.file)
  
  class(mcmc) <- "ppm.main"
  invisible(mcmc)
}


pre.process <- function(grid.data.file = NA, crime.data.file = NA,
                        num.kern.lon = 20, num.kern.lat = 20,
                        popden.scale = 1 / 30000, vacant.scale = 1,
                        domain)
{
  ## Process covariate data
  grid.data <- read.csv(grid.data.file, header = TRUE)
  corv <- data.frame(lon = grid.data$X, lat = grid.data$Y,
                     popden = (grid.data$POP00_SQMI * popden.scale), 
                     vacant = ((grid.data$VACANT / grid.data$HSE_UNITS) * vacant.scale), 
                     grid.area = (grid.data$CNTY * grid.data$SQMI))

  ## Discard covariates from other counties and set NaN vacancy rate to 0
  corv$popden[corv$CNTY == 0] <- 0
  corv$vacant[corv$CNTY == 0] <- 0
  corv$vacant[is.nan(corv$vacant)] <- 0
  
  crime.data <- read.csv(crime.data.file, header = TRUE)
  obsv <- data.frame(lon = crime.data$X, lat = crime.data$Y, 
                     weekday = crime.data$weekday,
                     month = crime.data$month, 
                     week = crime.data$week,
                     week.mid = floor((crime.data$weekday + 3) / 7) + crime.data$week, 
                     popden = (crime.data$POP00_SQMI * popden.scale), 
                     vacant = (crime.data$VACANT / crime.data$HSE_UNITS) * vacant.scale)
    
  ## Disgard data that did not not get geo-coded and set NaN vacancy rate to 0
  obsv <- obsv[(obsv$lon != 0), ]
  obsv <- obsv[(obsv$lat != 0), ]
  obsv$vacant[is.nan(obsv$vacant)] <- 0

  X <- sort(unique(grid.data$X))
  Y <- sort(unique(grid.data$Y))
  del.x <- diff(c(X[1], X[2]))
  del.y <- diff(c(Y[1], Y[2]))

  ## Determine number of basis points
  if (is.null(num.kern.lon))
    num.kern.lon <- 22
  else
    num.kern.lon <- num.kern.lon + 2
  if (is.null(num.kern.lat))
    num.kern.lat <- 22
  else
    num.kern.lat <- num.kern.lat + 2

  ## Construct Kernel Locations (place some kernels outside of the grid)
  kern.lon <- seq(min(X) - del.x, max(X) + del.x, length = num.kern.lon)
  kern.lat <- seq(min(Y) - del.y, max(Y) + del.y, length = num.kern.lat)
  kernel <- data.frame(lon = rep(unique(kern.lon), times = length(unique(kern.lat))), 
                       lat = rep(unique(kern.lat), each = length(unique(kern.lon))))
  delta.x <- diff(sort(unique(kern.lon))[1:2])
  delta.y <- diff(sort(unique(kern.lat))[1:2])
  return(list(corv = corv, obsv = obsv, kernel = kernel, delta = c(delta.x, delta.y), 
              num.kern = c(num.kern.lon, num.kern.lat), domain = domain))
}


ppmffbs <- function(y, m0, C0, alpha0, beta0, T, n, del.star, discount) {
  m.t <- m0
  C.t <- C0
  alpha.t <- alpha0
  beta.t <- beta0
  v2 <- (1 / rgamma(1, alpha0, beta0))
  
  y <- matrix(y, nrow = n, ncol = T)
  m <- matrix(NA, nrow = n, ncol = T)
  a <- matrix(NA, nrow = n, ncol = T)
  C <- matrix(NA, nrow = n, ncol = T)
  R <- matrix(NA, nrow = n, ncol = T)
  theta <- matrix(NA, nrow = n, ncol = T)
  
  for (t in 1:T) {
    a.t <- m.t
    R.t <- C.t / discount
    e.t <- y[, t] - a.t
    Q.t <- R.t + v2
    Q.t.inv <- 1 / Q.t
    A.t <- R.t * Q.t.inv
    m.t <- a.t + A.t * e.t
    C.t <- R.t - A.t^2 * Q.t
    alpha.t <- del.star * alpha.t + n /2
    beta.t <- del.star * beta.t + 0.5 * sum(e.t^2 * Q.t.inv)
    v2 <- (1 / rgamma(1, alpha.t, beta.t))
    
    m[, t] <- m.t
    C[, t] <- C.t
    a[, t] <- a.t
    R[, t] <- R.t
  }
  
  theta.s <- rnorm(n)
  theta[, T] <- theta.s * sqrt(C[, T]) + m[, T]
  
  for (t in (T-1):1) {
    B.t <- C[, t] * (1 / R[, (t + 1)])
    m.ts <- m[, t] + B.t * (theta[, (t + 1)] - a[, (t + 1)])
    C.ts <- C[, t] - B.t^2 * R[, (t + 1)]
    theta.s <- rnorm(n)
    theta[, t] <- theta.s * sqrt(C.ts) + m.ts
  }

  return(list(m.ff = as.vector(m), C.ff = as.vector(C),
              m.bs = as.vector(theta), m.t = m.t, C.t = C.t,
              alpha.t = alpha.t, beta.t = beta.t, v2 = v2,
              f = m.t))
}
