################################################################################
#
# Point process modeling via process convolutions
# Copyright (C) 2011, University of California
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Questions? Contact Waley W. Liang (wliang@ams.ucsc.edu)
#
################################################################################
  
library(fields, verbose = FALSE)
library(akima, verbose = FALSE)
library(PBSmapping, verbose = FALSE)

## zu.i has the same meaning as in plot.zu.each.month
trace.plot.samples <- 
  function(mcmc, item, zu.ind = 1:4, acf = TRUE, file = "", width = 5, height = 5, ...)
{
  if (mcmc$num.beta == 2)
    scale <- c(mcmc$popden.scale, 1)
  else
    scale <- c(1, mcmc$popden.scale, 1)
  if (item == "zu") {
    if (file == "")
      x11(width = 2*width, height = (if (acf) height else 0.5*height))
    else
      postscript(file = paste(file, "eps", sep = "."),
                 width = 2*width, height = (if (acf) height else 0.5*height))
    par(mfrow = c((if (acf) 2 else 1), 4))
    for (j in 1:length(zu.ind)) {
      t <- floor((zu.ind[j] - 1) / nrow(mcmc$data$kern)) + 1
      if (mcmc$dlm)
        main <- substitute(z[i](u[j]), list(i=t, j=zu.ind[j] %% nrow(mcmc$data$kern)))
      else
        main <- substitute(z(u[j]), list(j=zu.ind[j]))
      plot(mcmc$zu[zu.ind[j], ], cex = 2, xlab = "index", ylab = "posterior value",
           main = main, ...)
      legend("bottomright", legend = paste("AR: ",
                              round(mcmc$accept.zu[zu.ind[j]] * 100, 2), "%", sep=""))
      abline(h = mcmc$zu.mean[zu.ind[j]], col = "darkgray")
    }
    if (acf) {
      for (j in 1:length(zu.ind)) {
        t <- floor((zu.ind[j] - 1) / nrow(mcmc$data$kern)) + 1
        if (mcmc$dlm)
          main <- substitute(z[i](u[j]), list(i=t, j=zu.ind[j] %% nrow(mcmc$data$kern)))
        else
          main <- substitute(z(u[j]), list(j=zu.ind[j]))
        acf(mcmc$zu[zu.ind[j], ], xlab = "index", ylab = "acf", main = main)
      }
    }
  } else if (item == "tau") {
    if (file == "")
      x11(width = 2*width, height = height)
    else
      postscript(file = paste(file, "eps", sep = "."), width = 2*width, height = height)
    par(mfrow = c(1, 2), oma = c(2, 0, 2, 0))
    plot(mcmc$tau, cex = 2, xlab = "index", ylab = "posterior value", ...)
    abline(h = mean(mcmc$tau), col = "darkgray")
    if (acf)
      acf(mcmc$tau, xlab = "index", ylab = "acf", main = expression(paste("ACF of ", tau)))
    else {
      hist(mcmc$tau, main = "", xlab = "tau")
      abline(v = mcmc$tau.mean, col = "darkgray")
    }
    title(main = expression(paste("Posterior samples of ", tau)), outer = T)
  } else if (item == "beta") {
    if (file == "")
      x11(width = 1.5*width, height = (if (acf) height else 0.5*height))
    else
      postscript(file = paste(file, "eps", sep = "."), width = 1.5*width,
                 height = (if (acf) height else 0.5*height))
    par(mfrow = c((if (acf) 2 else 1), mcmc$num.beta))
    for (j in 1:mcmc$num.beta) {
      plot(mcmc$beta[j, ] * scale[j], cex = 2, xlab = "index",
           ylab = "posterior value", main = substitute(beta[j], list(j=j)), ...)
      legend("bottomright", legend = paste("AR: ",
                              round(mcmc$accept.beta[j] * 100, 2), "%", sep=""))
      abline(h = mcmc$beta.mean[j] * scale[j], col = "darkgray")
    }
    if (acf) {
      for (j in 1:mcmc$num.beta) {
        acf(mcmc$beta[j, ] * scale[j], xlab = "index", ylab = "acf",
            main = substitute(beta[j], list(j=j)))
      }
    }
  } else if (item == "phi") {
    if (file == "")
      x11(width = width, height = (if (acf) height else 0.5*height))
    else
      postscript(file = paste(file, "eps", sep = "."), width = width,
                 height = (if (acf) height else 0.5*height))
    fit.weekday <- mcmc$fit.weekday
    fit.month <- mcmc$fit.month
    if (fit.weekday & fit.month)
      num.plots <- 2
    else if (xor(fit.weekday, fit.month))
      num.plots <- 1
    else{
      stop("Time components not available")
    }
    par(mfrow = c((if (acf) 2 else 1), mcmc$num.phi))
    for (j in 1:mcmc$num.phi) {
      plot(mcmc$phi[j, ], cex = 2, xlab = "index", ylab = "posterior value",
           main = substitute(phi[j], list(j=j)), ...)
      legend("bottomright", legend = paste("AR: ",
                              round(mcmc$accept.phi[j] * 100, 2), "%", sep=""))
      abline(h = mcmc$phi.mean[j], col = "darkgray")
    }
    if (acf) {
      for (j in 1:mcmc$num.phi) {
        acf(mcmc$phi[j, ], xlab = "index", ylab = "acf",
            main = substitute(phi[j], list(j=j)))
      }
    }
  } else {
    stop("Invalid item input")
  }
  if (file != "")
    dev.off()
}


## Plot spatial effect 
plot.spatial.effect <-
  function(mcmc, file = "", width = 10, height = 5, ...)
{
  beta <- mcmc$beta
  num.beta <- mcmc$num.beta
  if (file == "")
    x11(width = width, height = height)
  else
    postscript(file = paste(file, "eps", sep = "."), width = width, height = height)
  if (num.beta == 2) {
    par(mfrow = c(1, num.beta))
    hist(exp(1000 * beta[1, ] * mcmc$popden.scale), xlab = "exp(1000 * beta1)", 
         main = "Factor for 1000 people / sqmi increase", col = 'lightgray')
    abline(v = mean(exp(1000 * beta[1, ] * mcmc$popden.scale)), lty = 2, col = "black")
    hist(exp(0.1 * beta[2, ]), xlab = "exp(0.1 * beta2)", 
         main = "Factor for 10% vacancy rate increase", col = 'lightgray')
    abline(v = mean(exp(0.1 * beta[2, ])), lty = 2, col = "black")
  } else if (num.beta == 3) {
    par(mfrow = c(1, num.beta))
    hist(exp(beta[1, ]), xlab = expression(paste("exp(", beta[1], ")")),
         main = "Baseline Factor", col = 'lightgray', ...)
    abline(v = mean(exp(beta[1, ])), col = "black", lty = 2, lwd = 2)
    hist(exp(1000 * beta[2, ] * mcmc$popden.scale), 
         xlab = expression(paste("exp(1000 ", beta[2], ")")), 
         main = "Factor for 1000 people / sqmi increase", 
         col = 'lightgray', ...)
    abline(v = mean(exp(1000 * beta[2, ] * mcmc$popden.scale)), 
           col = "black", lty = 2, lwd = 2)
    hist(exp(0.1 * beta[3, ]), xlab = expression(paste("exp(0.1 ", beta[3], ")")), 
         main = "Factor for 10% vacancy rate increase", 
         col = 'lightgray', ...)
    abline(v = mean(exp(0.1 * beta[3, ])), col = "black", lwd = 2)
  }
  if (file != "")
    dev.off()
}


## Plot temporal effect (weekday and / or monthly) with credible 90% interval
plot.time.effect <-
  function(mcmc, file = "", width = 5, height = 5)
{
  phi <- mcmc$phi
  fit.weekday <- mcmc$fit.weekday
  fit.month <- mcmc$fit.month
  weekday <- seq(1, 7, by = 1)
  month <- seq(1, mcmc$num.months, by = 1)
  weekday.effect <- NULL
  month.effect <- NULL
  if (fit.weekday & fit.month)
    num.plots <- 2
  else if (xor(fit.weekday, fit.month))
    num.plots <- 1
  else
    return()
  if (file == "")
    x11(width = width, height = height)
  else
    postscript(file = paste(file, "eps", sep = "."),
               width = width, height = height)
  par(mfrow = c(1, num.plots))
  if (fit.weekday) {
    for (i in 1:ncol(phi)) {
      weekday.effect <-
        cbind(weekday.effect, 
              exp(phi[1, i] * sin(2 * pi * weekday / 7) +
                  phi[2, i] * cos(2 * pi * weekday / 7)))
    }
    q <- apply(weekday.effect, 1, quantile, c(0.1, 0.9))
    plot(weekday, apply(weekday.effect, 1, mean), type = "l", 
         xlab = "day of the week(1 = sunday)", 
         ylab = "Seasonal factor", 
         main = "Seasonal Model Component(day of the week)", 
         ylim = c(min(q[1, ]), max(q[2, ])), col = "black")
    lines(weekday, q[1, ], lty = "dashed", col = "darkgray")
    lines(weekday, q[2, ], lty = "dashed", col = "darkgray")
  }
  if (fit.month) {
    num.months <- mcmc$num.months
    if (fit.weekday)
      index <- 3
    else
      index <- 1
    month.effect <- exp(outer(sin(2 * pi * month / 12), phi[index, ]) +
                        outer(cos(2 * pi * month / 12), phi[index + 1, ]))
    q <- apply(month.effect, 1, quantile, c(0.05, 0.95))
    plot(month, apply(month.effect, 1, mean), type = "l", 
         xlab = "month", ylab = "Seasonal factor", 
         main = "Seasonal (month of the year) effects", 
         ylim = c(min(q[1, ]), max(q[2, ])), col = "black")
    lines(month, q[1, ], lty = "dashed", col = "darkgray")
    lines(month, q[2, ], lty = "dashed", col = "darkgray")
    legend("bottom", legend = c("mean", "90% prob limits"),
           lty = c("solid", "dashed"), col = c("black", "darkgray"))
  }
  if (file != "")
    dev.off()
}


## Plots the estimated intensity using perspective plot and contour plot
plot.intensity <- 
  function(mcmc, theta = 45, phi = 20, ltheta = -120, nlevels = 15,
           main = NULL, spatial = FALSE, time = FALSE, t = 1,
           file = "", width = 10, height = 5, pred.over.grid = FALSE,
           plot = TRUE, num.lon = 40, num.lat = 40, zlim = c(1.5, 25),
           landscape = TRUE, plot.contour = TRUE, add.basis = FALSE)
{
  if (mcmc$dlm) {
    if (is.null(mcmc$predict.month)) {
      if (pred.over.grid) {
        intensity <- mcmc$baseline.pred[[t]]
        if (spatial) {
          intensity <- intensity * mcmc$spat.eff.pred
        }
      } else {
        intensity <- mcmc$baseline.fit[[t]]
        if (spatial) {
          intensity <- intensity * mcmc$spat.eff.fit[[t]]
        }
      }
    } else {
      if (t == mcmc$predict.month) {
        intensity <- mcmc$baseline.forecast
        if (spatial) {
          intensity <- intensity * mcmc$spat.eff.pred
        }
      } else {
        if (pred.over.grid) {
          intensity <- mcmc$baseline.pred[[t]]
          if (spatial) {
            intensity <- intensity * mcmc$spat.eff.pred
          }
        } else {
          intensity <- mcmc$baseline.fit[[t]]
          if (spatial) {
            intensity <- intensity * mcmc$spat.eff.fit[[t]]
          }
        } 
      }
    } 
  } else {
    if (pred.over.grid) {
      intensity <- mcmc$baseline.pred
      if (spatial) {
        intensity <- intensity * mcmc$spat.eff.pred
      }
    } else {
      intensity <- mcmc$baseline.fit
      if (spatial) {
        intensity <- intensity * mcmc$spat.eff.fit
      }
    }
  }
  if (time) {
    intensity <- t(t(intensity) * mcmc$time.eff[t, ])
  }

  intensity <- apply(intensity, 1, mean)
  obsv <- mcmc$data$obsv
  corv <- mcmc$data$corv
  if (mcmc$dlm & !is.null(mcmc$predict.month)) {
    if (mcmc$predict.month == t) {
      pred.over.grid <- TRUE
    }
  }
  
  if (pred.over.grid) {
    interpolate <- interp(corv$lon, corv$lat, intensity)
  } else {
    if (mcmc$dlm) {
      interpolate <- interp(obsv$lon[obsv$month == t],
                            obsv$lat[obsv$month == t],
                            intensity, duplicate = "mean", linear = TRUE)
    } else {
      interpolate <- interp(obsv$lon, obsv$lat,
                            intensity, duplicate = "mean", linear = TRUE)
    }
  }
  
  x = interpolate$x
  y = interpolate$y
  z = interpolate$z
  
  if (mcmc$dlm) {
    if (!identical(mcmc$predict.month, t)) {
      z.base.pred <- interp(corv$lon, corv$lat, apply(mcmc$baseline.pred[[t]], 1, mean))$z
    } else {
      z.base.pred <- interp(corv$lon, corv$lat, apply(mcmc$baseline.forecast, 1, mean))$z
    }
  }
  else {
    z.base.pred <- interp(corv$lon, corv$lat, apply(mcmc$baseline.pred, 1, mean))$z
  }
  zi <- z.base.pred[-1, -1] + z.base.pred[-1, -num.lat] +
        z.base.pred[-num.lon, -1] + z.base.pred[-num.lon, -num.lat]

  fcol <- gray.colors(128)[cut(zi, quantile(zi, seq(0, 1, len = 129), na.rm = TRUE), 
                               include.lowest = FALSE)]
  
  if (plot) {
    if (file == "") {
      x11(width = 5, height = 5)
    }
    else {
      postscript(file = paste(file, "_persp.eps", sep = ""),
                 width = width, height = height)
    }
    persp(x, y, z, theta = theta, phi = phi, ltheta = ltheta, main = "",
          col = fcol, shade = 0.4, border = "black", box = TRUE, xlab = "longitude",
          ylab = "latitude", zlim = zlim)
    if (file != "")
      dev.off()
    
    if (file == "") {
      x11(width = width, height = height)
    }
    else {
      postscript(file = paste(file, "_cont.eps", sep = ""),
                 width = width, height = height)
    }
    if (plot.contour) {
      cinmap(main = main)
      contour(x, y, z, nlevels = nlevels, xlab = "Longitude", ylab = "Latitude", 
              add = TRUE, lwd = 2)
      if (add.basis) 
        text(mcmc$data$kernel[, 1], mcmc$data$kernel[, 2],
             labels = as.character(1:nrow(mcmc$data$kernel)))
    }
    if (file != "")
      dev.off()
  }
  invisible(intensity)
}


## Matt Taddy's function for plotting a map of Cincinnati
cinmap <-
  function(main = "Cincinnati Crime Data", xlim = c(-84.65, -84.37), 
           ylim = c(39.09, 39.23),  col = gray.colors(128),
           map = "map1", ...)
{
  ## MAP 1
  if (map == "map1") {
    map1 <- read.table("cincinnati_map1.dat", header = TRUE)
    dim(map1)
    mx <- map1[, 1]
    my <- map1[, 2]
    mz <- map1[, -c(1, 2)]
    stretch_x <- 1.25
    stretch_y <- 1.00
    offset_x <- 0.00
    offset_y <- 0.00 
  }
  ## MAP 2
  if (map == "map2") {
    long <- read.table("long_map2.dat")
    lat <- read.table("lat_map2.dat")
    mz <- read.table("cincinnati_map2.dat")
    mx <- long[, 1] 
    my <- lat[, 1]
    stretch_x <- 0.95
    stretch_y <- 1.15
    offset_x <- 0.01
    offset_y <- -0.005
  }
  rx1 <- mean(mx) - stretch_x * (mean(mx) - min(mx)) + offset_x
  rx2 <- mean(mx) + stretch_x * (mean(mx) - min(mx)) + offset_x
  ry1 <- mean(my) - stretch_y * (mean(my) - min(my)) + offset_y
  ry2 <- mean(my) + stretch_y * (mean(my) - min(my)) + offset_y
  rmx <- seq(rx1, rx2, length = length(mx))
  rmy <- seq(ry1, ry2, length = length(my))
  mapstretch <- as.matrix(mz)
  image(rmx, rmy, mapstretch, main = main, col = col, 
        ylim = ylim , xlim = xlim, xlab = "Longitude", ylab = "Latitude", ...)
}
 

## Plot observation window
plot.obsv.window <-
  function(mcmc, plot.obsv = TRUE, plot.grid = FALSE, main = NULL, 
           plot.kern = FALSE, regular = TRUE, file = "", width = 5, height = 5)
{
  if (file == "")
    x11(width = width, height = height)
  else
    postscript(file = paste(file, "eps", sep = "."),
               width = width, height = height)
  
  if (plot.grid) {
    if (regular) {
      corv <- mcmc$data$corv
      lon <- sort(unique(corv$lon))
      lat <- sort(unique(corv$lat))
      del.x <- diff(lon[1:2])
      del.y <- diff(lat[1:2])
      lon <- lon + del.x / 2
      lon <- c(min(lon) - del.x, lon)
      lat <- lat + del.y / 2
      lat <- c(min(lat) - del.y, lat)
      for (i in 1:length(lon)) {
        segments(lon[i], min(lat), lon[i], max(lat))
      }
      for (i in 1:length(lat)) {
        segments(min(lon), lat[i], max(lon), lat[i])
      }
    } else { 
      require("PBSmapping")
      PopPoly <- read.csv("popPBSpoly.csv")
      plotPolys(PopPoly, main = main)
      cinmap(add = TRUE)
      addPolys(PopPoly)
    }
  } else {
    cinmap(main = main)
  }

  if (plot.obsv) {
    points(x = mcmc$data$obsv$lon, y = mcmc$data$obsv$lat,
           pch = 46, col = "black", cex = 2.5)
  }
  if (plot.kern) {
    points(x = mcmc$data$kern$lon, y = mcmc$data$kern$lat,
           pch = 19, col = "black", cex = 0.5)
  }

  if (file != "")
    dev.off()
}


## Returns TRUE if point (px, py) is within the
## ellipse with center (ex, ey) and radiants
## ea and eb, else FALSE
point.in.ellipse <-
  function(ex, ey, ea, eb, px, py)
{
  dx <- px - ex
  dy <- py - ey
  dist <- sqrt(dx^2 + dy^2)
  in.ellipse <- ((dx / ea)^2 + (dy / eb)^2) < 1
  return(list(dist = dist, in.ellipse = in.ellipse))
}


## Plot values of zu and number of observations around zu v.s. month
plot.zu.vs.month.dlm <-
  function(mcmc, j, file = "", width = 10, height = 5)
{
  num.months <- mcmc$num.months
  zu <- matrix(mcmc$zu.mean, nrow(mcmc$data$kern), num.months)
  count <- rep(0, num.months)
  ex <- mcmc$data$kernel$lon[j]
  ey <- mcmc$data$kernel$lat[j]
  for (t in 1:num.months) {
    obsv <- mcmc$data$obsv[mcmc$data$obsv$month==t, ]
    for (i in 1:nrow(obsv)) {
      point <- point.in.ellipse(ex, ey, mcmc$data$delta[1], mcmc$data$delta[2], 
                                obsv$lon[i], obsv$lat[i])
      if (point$in.ellipse) {
        count[t] = count[t] + 1
      }
    }
  }
  if (file == "")
    x11(width = width, height = height)
  else
    postscript(file = paste(file, "eps", sep = "."))
  par(oma = c(0, 2, 0, 2))
  TwoVarPlot(seq(1, num.months, 1), zu[j, ], count, 
             labels = c("t (month)", paste("E[ z_t(u_", j, ") | D_t ]", sep = ""),
             "observed count"), title = paste("u_", j, sep = ""),
             legpos = "topright")
  if (file != "")
    dev.off()
}


## Plots an x y1 y2 using left and right axes for the different y's
## rescales y2 to fit in the same space as y1
TwoVarPlot <-
  function(xvar, y1var, y2var, labels, noincs = 5, marks = c(1, 2),
           legpos, leglabs, title)
{
  ## Rescale to fit on same axis
  scaledy2var <- (y2var - min(y2var)) / (max(y2var) - min(y2var))
  scaledy2var <- (scaledy2var * (max(y1var) - min(y1var))) + min(y1var)

  ## Plot it up and add the points
  plot(xvar, y1var, xlab = labels[1], ylab = "", axes = F, pch = marks[1], 
       main = title, type = "l", col = "black")
  lines(xvar, scaledy2var, pch = marks[2], col = "darkgray", lty = "dashed")
  
  ## Make up some labels and positions
  y1labs <- round(seq(min(y1var), max(y1var), length = noincs), 4)
  
  ## Convert these to the y2 axis scaling
  y2labs <- (y1labs - min(y1var)) / (max(y1var) - min(y1var))
  y2labs <- (y2labs * (max(y2var) - min(y2var))) + min(y2var)
  y2labs <- round(y2labs, 2)

  axis(1)
  axis(2, at = y1labs, labels = y1labs)
  axis(4, at = y1labs, labels = y2labs)
  mtext(labels[3], side = 4, line = 2)
  mtext(labels[2], side = 2, line = 2)
  box()

  legend(legpos, legend = c(labels[2], labels[3]), col = c("black", "darkgray"), 
         lty = c(1, 2), lwd = 1)
}


######################### Anomaly Detection #############################

## Find out which point in "points" are within the circcular window
## with radius "radius" centered at "centroid"
## "centroid" is a vector of length 2
## "points" is a n by 2 matrix where each row is a point
## "radius" is the radius of the circular window
## Returns the indices of the points that are in the circular window
is.in.circle <- function(centroid, points, radius) {
  dist <- sqrt(rowSums(t(t(points) - centroid)^2))
  return(which(dist <= radius))
}


## Find out which grid cell in "grid.pts" contains
## observation "case.locs"
## Returns the index of the grid cell
which.grid <- function(case.locs, grid.pts) {
  dist <- sqrt(rowSums(t(t(grid.pts) - case.locs)^2))
  return(which(dist == min(dist)))
}

## Find the number of observations within each grid cell
## Returns a vector of length = number of grid cells containing
## the number of observations
find.grid.num.obsv <- function(grid, obsv, regular = TRUE) {
  n <- nrow(obsv)
  num.obsv <- rep(0, nrow(grid))
  if (regular) {
    for (i in 1:n) {
      ind <- which.grid(c(obsv$lon[i], obsv$lat[i]), cbind(grid$lon, grid$lat))[1]
      num.obsv[ind] <- num.obsv[ind] + 1
    }
  } else {
    for (i in 1:nrow(grid)) {
      num.obsv[i] <- length(which(grid$PID[i] == obsv$PID))
    }
  }
  return(num.obsv)
}

## Compute the expected number of observation within each grid cell
## Returns a vector of length = # of grid cells containing the
## expected number of observations
expected.num.obsv <- function(mcmc, month = 1, dlm = TRUE) {
  if (dlm) {
    if (is.null(mcmc$predict.month)) {
      baseline.pred <- mcmc$baseline.pred[[month]]
    }
    else if (month != mcmc$predict.month) {
      baseline.pred <- mcmc$baseline.pred[[month]]
    }
    else {
      baseline.pred <- mcmc$baseline.forecast
    }
  }
  else
    baseline.pred <- mcmc$baseline.pred
  grid.area <- mcmc$data$corv$grid.area
  spatial.effect <- exp(mcmc$Xs.grid %*% mcmc$beta.mean)
  time.effect <- exp(mcmc$time$Ft.cal %*% mcmc$phi.mean)[month]
  mean.num.obsv <- baseline.pred * spatial.effect * time.effect * grid.area
  return(mean.num.obsv)
}


## Compute and plot spatial scan results
spat.scan <- function(mcmc, month = NULL, radius, num.cluster = 3,
                      regular, label = NULL, file = "")
{
  corv <- mcmc$data$corv
  obsv <- mcmc$data$obsv
  XY <- cbind(corv$lon, corv$lat)
  num.months <- mcmc$num.months
  predict.month <- mcmc$predict.month
  num.kern <- mcmc$num.kern
  grid.area <- mcmc$data$corv$grid.area
  nobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
  eobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
  eobsv.mcmc <- array(NA, dim = c(num.months, nrow(corv), mcmc$iter))
  baseline.pred <- mcmc$baseline.pred
  
  if (mcmc$dlm) {
    if (!is.null(mcmc$predict.month)) {
      baseline.pred <- c(baseline.pred, list(mcmc$baseline.forecast))
      num.months <- num.months + 1
      nobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
      eobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
      eobsv.mcmc <- array(NA, dim = c(num.months, nrow(corv), mcmc$iter))
    }
    for (t in 1:num.months) {
      intensity <- baseline.pred[[t]] * mcmc$spat.eff.pred
      intensity <- t(t(intensity) * mcmc$time.eff[t, ])
      eobsv.mcmc[t, , ] <- intensity * grid.area
      eobsv[t, ] <- apply((intensity * grid.area), 1, mean)
      if (!regular) {
        corv <- read.csv("polycentresData.csv")
        obsv <- read.csv("crimeclass1points.csv")
        XY <- cbind(corv$X, corv$Y)
      }
      nobsv[t, ] <- find.grid.num.obsv(corv, obsv[obsv$month == t, ], regular)      
    }
  } else {
    for (t in 1:num.months) {
      intensity <- baseline.pred * mcmc$spat.eff.pred
      intensity <- t(t(intensity) * mcmc$time.eff[t, ])
      eobsv.mcmc[t, , ] <- intensity * grid.area
      eobsv[t, ] <- apply((intensity * grid.area), 1, mean)
    }
    if (!regular) {
      corv <- read.csv("polycentresData.csv")
      obsv <- read.csv("crimeclass1points.csv")
      XY <- cbind(corv$X, corv$Y) 
    }
    nobsv[t, ] <- find.grid.num.obsv(corv, obsv[obsv$month == t, ], regular)      
  }
  
  nn <- nrow(corv)
  windows <- vector("list", nn)
  test.stat <- rep(0, nn)
  cluster <- rep(0, nn)
  if (is.null(month)) {
    t.ind <- 1:nrow(nobsv)
    main <- paste("Spatial scanned clusters over all months")
  } else {
    t.ind <- month
    main <- paste("Spatial scanned clusters for month", month)
  }
  for (i in 1:nn) {
    ind <- is.in.circle(XY[i, ], XY, radius)
    windows[[i]] <- ind
    nz <- sum(nobsv[t.ind, ind])
    xiz <- sum(eobsv[t.ind, ind])
    nzc <- sum(nobsv[t.ind, ]) - nz
    xizc <- sum(eobsv[t.ind, ]) - xiz
    
    if (!is.nan(nz / xiz) & !is.nan(nzc / xizc)) { 
      if (nz / xiz > nzc / xizc) {
        test.stat[i] <- (nz * (log(nz) - log(xiz)) +
                         nzc * (log(nzc) - log(xizc))) ## on the log scale
      }
    } else {
      test.stat[i] <- 0
    }
  }
  
  mlcs <- sort(unique(test.stat), decreasing = TRUE)
  for (k in 1:num.cluster) {
    wind.ind <- windows[[which(test.stat == mlcs[k])[1]]]
    cluster[wind.ind] <- cluster[wind.ind] + 1
  }
  nz.cluster <- cluster[cluster > 0]
  st.cluster <- sort(cluster, decreasing = TRUE, index.return = TRUE)
  st.cluster.ind <- st.cluster$ix
  colors <- gray.colors(length(nz.cluster))
  col <- rep("white", nn)
  col[st.cluster.ind] <- c(colors, rep("white", (nn - length(nz.cluster))))
  require("PBSmapping")
  PopPoly <- read.csv("popPBSpoly.csv")
  if (file != "")
    postscript(file = file, width = 6.5, height = 5)
  else
    x11(width = 6.5, height = 5)
  plotPolys(PopPoly, col = col, main = main, xlab = "Longitude",
            ylab = "Latitude")
  if (!is.null(label)) {
    text(mcmc$data$corv$lon[label], mcmc$data$corv$lat[label],
         labels = as.character(label), cex = 0.75, col = "white")
  }
  if (file != "")
    dev.off()
}


detect.anomaly <- function(mcmc, regular = TRUE) {
  corv <- mcmc$data$corv
  obsv <- mcmc$data$obsv
  num.months <- mcmc$num.months
  predict.month <- mcmc$predict.month
  num.kern <- mcmc$num.kern
  grid.area <- mcmc$data$corv$grid.area
  nobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
  eobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
  eobsv.mcmc <- array(NA, dim = c(num.months, nrow(corv), mcmc$iter))
  baseline.pred <- mcmc$baseline.pred
  
  if (mcmc$dlm) {
    if (!is.null(mcmc$predict.month)) {
      baseline.pred <- c(baseline.pred, list(mcmc$baseline.forecast))
      num.months <- num.months + 1
      nobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
      eobsv <- matrix(0, nrow = num.months, ncol = nrow(corv))
      eobsv.mcmc <- array(NA, dim = c(num.months, nrow(corv), mcmc$iter))
    }
    for (t in 1:num.months) {
      intensity <- baseline.pred[[t]] * mcmc$spat.eff.pred
      intensity <- t(t(intensity) * mcmc$time.eff[t, ])
      eobsv.mcmc[t, , ] <- intensity * grid.area
      eobsv[t, ] <- apply((intensity * grid.area), 1, mean)
      if (!regular) {
        corv <- read.csv("polycentresData.csv")
        obsv <- read.csv("crimeclass1points.csv")
      }
      nobsv[t, ] <- find.grid.num.obsv(corv, obsv[obsv$month == t, ], regular)      
    }
  } else {
    for (t in 1:num.months) {
      intensity <- baseline.pred * mcmc$spat.eff.pred
      intensity <- t(t(intensity) * mcmc$time.eff[t, ])
      eobsv.mcmc[t, , ] <- intensity * grid.area
      eobsv[t, ] <- apply((intensity * grid.area), 1, mean)
    }
    if (!regular) {
      corv <- read.csv("polycentresData.csv")
      obsv <- read.csv("crimeclass1points.csv")
    }
    nobsv[t, ] <- find.grid.num.obsv(corv, obsv[obsv$month == t, ], regular)      
  }

  ## Spatial anomalies
  nobsv.sum <- apply(nobsv, 2, sum)
  eobsv.sum <- apply(eobsv, 2, sum)  
  spatial.p.value <- numeric(nrow(corv))
  for (i in 1:length(spatial.p.value)) {
    spatial.p.value[i] <- (1 - ppois(nobsv.sum[i], eobsv.sum[i])) +
                          dpois(nobsv.sum[i], eobsv.sum[i]) 
  }
  ind <- which((nobsv.sum == 0) & (eobsv.sum == 0))
  spatial.p.value[ind] <- 1
  
  if (regular) {
    spatial.p.value <- matrix(spatial.p.value, nrow = length(unique(corv$lon)),
                              ncol = length(unique(corv$lat)))
  }
  
  ## Temporal anomalies
  temporal.p.value <- matrix(0, num.months, nrow(corv))
  for (t in 1:num.months) {
    for (i in 1:nrow(corv)) {
      temporal.p.value[t, i] <- (1 - ppois(nobsv[t, i], eobsv[t, i])) +
                                dpois(nobsv[t, i], eobsv[t, i])
    }
    ind <- which((nobsv[t, ] == 0) & (eobsv[t, ] == 0))
    temporal.p.value[t, ind] <- 1
  }
  if (regular) {
    temporal.p.value <- array(temporal.p.value,
                              dim = c(num.months, length(unique(corv$lon)),
                                length(unique(corv$lat))))
  }
  
  mcmc <- c(mcmc, list(spatial.p.value = spatial.p.value,
                       temporal.p.value = temporal.p.value,
                       nobsv = nobsv, eobsv = eobsv, nobsv.sum = nobsv.sum,
                       eobsv.sum = eobsv.sum, eobsv.mcmc = eobsv.mcmc))
  return(mcmc)
}


plot.spatial.p.value <-
  function(mcmc, regular = FALSE, label = NULL, width = 6.5, height = 5, file = "")
{
  if (regular) {
    corv <- mcmc$data$corv
    x <- sort(unique(corv$lon))
    y <- sort(unique(corv$lat))
    spatial.p.value <- mcmc$spatial.p.value
    spatial.p.value[spatial.p.value == 0] <- 1
    spatial.p.value[spatial.p.value > 0.1] <- NA
    if (file != "")
      postscript(file = file, width = width, height = height)
    else
      x11(width = 5, height = 5)
    image(x = x, y = y, z = spatial.p.value, 
          breaks = c(0, 0.01, 0.05, 0.1), 
          col = rev(gray.colors(3)),
          main = "EP-based clusters over all months",
          xlab = "Longitude", ylab = "Latitude", add = FALSE)
    if (file != "")
      dev.off()
  } else {
    spatial.p.value <- mcmc$spatial.p.value
    sig.ep <- spatial.p.value[spatial.p.value < 0.1]
    colors <- gray.colors(length(sig.ep))
    col <- rep("white", length(spatial.p.value))
    st.ep <- sort(spatial.p.value, decreasing = FALSE, index.return = TRUE)
    col[st.ep$ix] <- c(colors, rep("white", (length(spatial.p.value) - length(sig.ep))))
    require("PBSmapping")
    PopPoly <- read.csv("popPBSpoly.csv")
    if (file != "")
      postscript(file = file, width = 6.5, height = 5)
    else
      x11(width = 6.5, height = 5)
    plotPolys(PopPoly, col = col, main = "EP-based clusters over all months",
              xlab = "Longitude", ylab = "Latitude")
    if (!is.null(label)) {
      text(mcmc$data$corv$lon[label], mcmc$data$corv$lat[label],
           labels = as.character(label), cex = 0.75, col = "white")
    }
    if (file != "")
      dev.off()
    return(length(which(col != "white")))
  }
}


plot.temporal.p.value <-
  function(mcmc, t = 12, regular = FALSE, label = c(48, 100),
           width = 6.5, height = 5, file = "")
{
  if (is.null(mcmc$predict.month))
    main <- paste("EP-based clusters for month", t)
  else if (t != mcmc$predict.month)
    main <- paste("EP-based clusters for month", t)
  else
    main <- paste("Prediction of EP-based clusters for month", t)
      
  if (regular) {
    corv <- mcmc$data$corv
    x <- sort(unique(corv$lon))
    y <- sort(unique(corv$lat))
    temporal.p.value <- mcmc$temporal.p.value
    temporal.p.value[temporal.p.value == 0] <- 1
    temporal.p.value[temporal.p.value > 0.1] <- NA
    if (file != "")
      postscript(file = file, width = width, height = height)
    else
      x11(width = 5, height = 5)
    image(x = x, y = y, z = temporal.p.value[t, , ],
          main = main, breaks = c(0, 0.01, 0.05, 0.1), 
          col = rev(gray.colors(3)), xlab = "Longitude",
          ylab = "Latitude", add = FALSE)
    if (file != "")
      dev.off()
  } else {
    temporal.p.value <- mcmc$temporal.p.value[t, ]
    sig.ep <- temporal.p.value[temporal.p.value < 0.1]
    colors <- gray.colors(length(sig.ep))
    col <- rep("white", length(temporal.p.value))
    st.ep <- sort(temporal.p.value, decreasing = FALSE, index.return = TRUE)
    col[st.ep$ix] <- c(colors, rep("white", (length(temporal.p.value) - length(sig.ep))))
    require("PBSmapping")
    PopPoly <- read.csv("popPBSpoly.csv")
    if (file != "")
      postscript(file = file, width = 6.5, height = 5)
    else
      x11(width = 6.5, height = 5)
    plotPolys(PopPoly, col = col, main = main,
              xlab = "Longitude", ylab = "Latitude")
    if (!is.null(label)) {
      text(mcmc$data$corv$lon[label], mcmc$data$corv$lat[label],
           labels = as.character(label), cex = 0.75, col = "white")
    }
    if (file != "")
      dev.off()
    return(length(which(col != "white")))
  }
}


plot.hist.ep <- function(mcmc, breaks = 30, file = "") {
  if (file != "")
    postscript(file = file, width = 5, height = 5)
  else
    x11(width = 5, height = 5)
  p.value <- as.vector((mcmc$temporal.p.value))
  prop <- length(which(p.value < 0.1)) / length(p.value < 0.1)
  hist(p.value, breaks = breaks, xlab="EP",
       main = "Histogram of EPs over all months", col = "lightgray")
  if (file != "")
    dev.off()
  return(prop)
}


plot.spatial.anomaly <- function(mcmc, grid.ind = NULL, file = "") {
  for (i in 1:nrow(mcmc$data$corv)) {
    if (!is.null(grid.ind)) {
      i <- grid.ind
    }
    nobsv <- mcmc$nobsv[, i]
    eobsv.quan <- apply(mcmc$eobsv.mcmc[, , i], 1, quantile,
                        probs = c(.025, .975))
    eobsv.mean <- apply(mcmc$eobsv.mcmc[, , i], 1, mean)
    if (length(which(nobsv > eobsv.quan[2, ])) == 12) {
      ylim <- range(c(nobsv, range(eobsv.quan), eobsv.mean))
      if (file != "")
        postscript(file = file, width = 5, height = 5)
      else
        x11(width = 5, height = 5)
      plot(nobsv, ylim = ylim, main = paste("Region", i, sep = " "),
           xlab = "time", ylab = "counts", col = "black", pch = 19,
           cex = 0.7)
      lines(eobsv.quan[1, ], lty = "dashed", col = "darkgray")
      lines(eobsv.quan[2, ], lty = "dashed", col = "darkgray")
      lines(eobsv.mean, col = "black")
      if (file != "")
        dev.off()
    }
    if (!is.null(grid.ind)) {
      break
    }
  }
}


plot.temporal.anomaly <- function(mcmc, grid.ind = NULL, file = "") {
  for (i in 1:nrow(mcmc$data$corv)) {
    if (!is.null(grid.ind)) {
      i <- grid.ind
    }
    nobsv <- mcmc$nobsv[, i]
    eobsv.quan <- apply(mcmc$eobsv.mcmc[, , i], 1, quantile,
                        probs = c(.025, .975))
    eobsv.mean <- apply(mcmc$eobsv.mcmc[, , i], 1, mean)
    if (length(which(nobsv > eobsv.quan[2, ])) == 1) {
      ylim <- range(c(nobsv, range(eobsv.quan), eobsv.mean))
      if (file != "")
        postscript(file = file, width = 5, height = 5)
      else
        x11(width = 5, height = 5)
      plot(nobsv, ylim = ylim, main = paste("Region", i, sep = " "),
           xlab = "time", ylab = "counts", col = "black", pch = 19,
           cex = 0.7)
      lines(eobsv.quan[1, ], lty = "dashed", col = "darkgray")
      lines(eobsv.quan[2, ], lty = "dashed", col = "darkgray")
      lines(eobsv.mean, col = "black")
      if (file != "")
        dev.off()
    }
    if (!is.null(grid.ind)) {
      break
    }
  }
}


## Make Q-Q plot of residuals
plot.qq <- function(mcmc, less.sig = FALSE, alpha = 0.01, pearson = FALSE,
                    main, width = 5, height = 5, file = NULL, ppdf = TRUE) {
  T <- 12
  eobsv.mcmc <- mcmc$eobsv.mcmc
  temporal.p.value <- mcmc$temporal.p.value
  month <- c("January", "February", "March", "April", "May", "June",
             "July", "August", "September", "October", "November", "December")
  if (is.null(file)) {
    x11(width = width, height = height)
  } else {
    if (ppdf)
      pdf(file = file, width = width, height = height)
    else
      postscript(file = file, width = width, height = height)
  }
  par(mfrow = c(2, 6), oma = c(2, 0, 2, 0))
  for (t in 1:T) {
    nobsv <- mcmc$nobsv[t, ]
    eobsv <- eobsv.mcmc[t, , ]
    insig.ind <- 1:ncol(eobsv)
    if (less.sig) {
      insig.ind <- which(temporal.p.value[t, ] >= alpha)
      nobsv <- nobsv[insig.ind]
      eobsv <- eobsv[insig.ind, ]
    }
    n.sim <- ncol(eobsv)
    n.locs <- nrow(eobsv)
    sim <- matrix(NA, nrow = n.locs, ncol = n.sim)
    for (i in 1:n.sim) {
      sim[, i] <- rpois(n.locs, eobsv[, i])
    }
    sim.resid <- (sim - eobsv)
    resid <- (nobsv - eobsv)
    if (pearson) {
      sim.resid <- sim.resid / sqrt(eobsv)
      resid <- resid / sqrt(eobsv)
    }
    sim.resid <- apply(sim.resid, 2, sort)
    sim.resid.mu <- apply(sim.resid, 1, mean)
    sim.resid.up <- apply(sim.resid, 1, quantile, 0.975)
    sim.resid.lo <- apply(sim.resid, 1, quantile, 0.025)      
    resid.mu <- sort(apply(resid, 1, mean))
    xylim <- range(c(sim.resid.mu, resid.mu))
    plot(sim.resid.mu, resid.mu, xlab = "Expected quantiles",
         ylab = "Empirical quantiles", main = month[t], xlim = xylim, ylim = xylim)
    lines(sim.resid.mu, sim.resid.up, col = "darkgray", lty = "dashed")
    lines(sim.resid.mu, sim.resid.lo, col = "darkgray", lty = "dashed")
    abline(a = 0, b = 1)
    ##legend("bottomright", legend = c("95% interval"),
    ##       lty = c("dashed"), col = c("darkgray"))
  }
  title(main = main, outer = T)
  if (!is.null(file))
    dev.off()
}
