################################################################################
#
# Point process modeling via process convolutions
# Copyright (C) 2011, University of California
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Questions? Contact Waley W. Liang (wliang@ams.ucsc.edu)
#
################################################################################

source("ppm.main.R")
source("ppm.plots.R")

makeplots <- FALSE
print <- FALSE
poster <- FALSE
output.file.JanToDec <- "ppm.dlm.JanToDec.RData"

## Run dynamic model using all data
set.seed(10)
runtime <-
  system.time(mcmc <-
              main(domain = cbind(c(-84.65, -84.37), c(39.09, 39.23)),
                   grid.data.file = "polycentresData.csv",
                   crime.data.file = "crimeclass1points.csv", 
                   burnin = 1000, iter = 2000, proposal = 2,
                   kernel.fun = "bezier", kernel.w = 3, kernel.l = 3,
                   num.kern.lon = 40, num.kern.lat = 20,
                   pro.zu.a = 15, pro.zu.del = 0.5, dlm = TRUE,
                   pri.zu.shape = 1, pri.zu.rate = 1,
                   pri.beta.mean = c(0, 0), pri.beta.sd = c(1, 1),
                   pri.phi.sd = c(1, 1), pro.beta.sd = c(0.15, 0.25),
                   pro.phi.sd = c(0.1, 0.1), tau = (3857 / (12 * 138)),
                   zu.init = 0.001, beta.init = c(0, 0),
                   phi.init = c(0, 0, 0, 0), v = 0.1, m0 = log(0.001), c0 = 1,
                   alpha0 = 1, beta0 = 0.001, del.star = 0.95, discount = 0.95,
                   predict.month = NULL, output.file = output.file.JanToDec))

if (makeplots) {
  load(output.file.JanToDec)
  
  ## Plot spatial effects of population density and vacancy rates
  plot.spatial.effect(mcmc, file = (if (print) "spatial_effect_d" else ""), breaks = 10)
  
  ## Plot monthly effect
  plot.time.effect(mcmc, file = (if (print) "time_effect_d" else ""))
  
  ## Plot intensity for Feb.
  plot.intensity(mcmc, theta = 30, nlevels = 20, spatial = T, time = T, t = 2,
                 file = (if (print) "intensity_feb_d" else ""), zlim = c(0, 200),
                 landscape = (if (poster) FALSE else TRUE),
                 width = 6.5, height = 5, main = "Fitted intensity for Febuary")
  
  ## Plot intensity for June
  plot.intensity(mcmc, theta = 30, nlevels = 20, spatial = T, time = T, t = 6,
                 file = (if (print) "intensity_jun_d" else ""), zlim = c(0, 200),
                 landscape = (if (poster) FALSE else TRUE),
                 width = 6.5, height = 5, main = "Fitted intensity for June")
  
  ## Plot intensity for Dec.
  plot.intensity(mcmc, theta = 30, nlevels = 20, spatial = T, time = T, t = 12,
                 file = (if (print) "intensity_dec_d" else ""), zlim = c(0, 200),
                 landscape = (if (poster) FALSE else TRUE),
                 width = 6.5, height = 5, main = "Fitted intensity for December")
  
  ## Plot E[ z_t(u_j) | D_t ] and # of observations within the support of the
  ## bezier kernel centered at u_j, v.s. months
  plot.zu.vs.month.dlm(mcmc, 105, file = (if (print) "zu105_vs_month_d" else ""))
  plot.zu.vs.month.dlm(mcmc, 147, file = (if (print) "zu147_vs_month_d" else ""))
  plot.zu.vs.month.dlm(mcmc, 189, file = (if (print) "zu189_vs_month_d" else ""))
  plot.zu.vs.month.dlm(mcmc, 231, file = (if (print) "zu231_vs_month_d" else ""))

  ## Anomaly detection
  mcmc <- detect.anomaly(mcmc, regular = FALSE)

  ## Q-Q plots
  ## Run detect.anomaly first
  plot.qq(mcmc, less.sig = FALSE, alpha = 0.01, pearson = FALSE,
          main = "Empirical quantiles of raw residuals v.s. expected quantiles of simulated raw residuals for the dynamic model with full data",
          width = 10, height = 5, file = if (print) "qqplot_raw_resid.eps", ppdf = FALSE)
  plot.qq(mcmc, less.sig = FALSE, alpha = 0.01, pearson = TRUE,
          main = "Empirical quantiles of Pearson residuals v.s. expected quantiles of simulated Pearson residuals for the dynamic model with full data",
          width = 10, height = 5, file = if (print) "qqplot_pearson_resid.eps", ppdf = FALSE)
  
  ## EP-based clusters over all months
  nspt <- plot.spatial.p.value(mcmc, label = c(41, 48, 100, 182), width = 6.5, height = 5,
                               file = (if (print) "dlm_all_clusters.eps" else ""))
  
  ## EP-based clusters for Dec
  ntmp <- plot.temporal.p.value(mcmc, t = 12, label = c(41, 48, 100, 182), width = 6.5,
                                height = 5, file = (if (print) "dlm_dec_clusters.eps" else ""))

  ## Spatial scanned clusters over all months
  spat.scan(mcmc, month = NULL, 0.00089, num.cluster = 79, regular = FALSE,
            label = c(41, 48, 100, 182), file = (if (print) "dlm_spat_scan_all_clusters.eps" else ""))

  ## Spatial scanned clusers for Dec
  spat.scan(mcmc, month = 12, 0.00089, num.cluster = 19, regular = FALSE,
            label = c(41, 48, 100, 182),file = (if (print) "dlm_spat_scan_dec_clusters.eps" else ""))

  plot.spatial.anomaly(mcmc, file = (if (print) "dlm_spat_anomaly.eps" else ""))
  
  plot.temporal.anomaly(mcmc, grid.in = 48, file = (if (print) "dlm_temp_anomaly.eps" else ""))
  
  prop <- plot.hist.ep(mcmc, file = (if (print) "dlm_histEP.eps" else ""))
}



################### Prediction for Dec using data from Jan to Nov ####################
makeplots <- FALSE
print <- FALSE
poster <- FALSE
output.file.JanToNov <- "ppm.dlm.JanToNov.RData"

## Run dynamic model using data from Jan to Nov
set.seed(10)
runtime <-
  system.time(mcmc <-
              main(domain = cbind(c(-84.65, -84.37), c(39.09, 39.23)),
                   grid.data.file = "polycentresData.csv",
                   crime.data.file = "crimeclass1points.csv", 
                   burnin = 1000, iter = 2000, proposal = 2,
                   kernel.fun = "bezier", kernel.w = 3, kernel.l = 3,
                   num.kern.lon = 40, num.kern.lat = 20,
                   pro.zu.a = 15, pro.zu.del = 0.5, dlm = TRUE,
                   pri.zu.shape = 1, pri.zu.rate = 1,
                   pri.beta.mean = c(0, 0), pri.beta.sd = c(1, 1),
                   pri.phi.sd = c(1, 1), pro.beta.sd = c(0.15, 0.25),
                   pro.phi.sd = c(0.1, 0.1), tau = (3857 / (12 * 138)),
                   zu.init = 0.001, beta.init = c(0, 0),
                   phi.init = c(0, 0, 0, 0), v = 0.1, m0 = log(0.001), c0 = 1,
                   alpha0 = 1, beta0 = 0.001, del.star = 0.95, discount = 0.95,
                   predict.month = 12, output.file = output.file.JanToNov))

if (makeplots) {
  load(output.file.JanToNov)
  
  plot.intensity(mcmc, theta = 30, nlevels = 20, spatial = T, time = T, t = 2,
                 file = (if (print) "intensity_feb_dp12" else ""),
                 main = "Fitted intensity for Febuary", width = 6.5, height = 5,
                 zlim = c(0, 200), landscape = (if (poster) FALSE else TRUE))
  
  ## Plot intensity for June
  plot.intensity(mcmc, theta = 30, nlevels = 20, spatial = T, time = T, t = 6,
                 file = (if (print) "intensity_jun_dp12" else ""),
                 main = "Fitted intensity for June", width = 6.5, height = 5,
                 zlim = c(0, 200), landscape = (if (poster) FALSE else TRUE))
  
  ## Plot intensity for Dec.
  plot.intensity(mcmc, theta = 30, nlevels = 20, spatial = T, time = T, t = 12,
                 file = (if (print) "intensity_dec_dp12" else ""),
                 main = "Forecasted intensity for December", width = 6.5, height = 5,
                 zlim = c(0, 200), landscape = (if (poster) FALSE else TRUE))
  
  ## Anomaly detection
  mcmc <- detect.anomaly(mcmc, regular = FALSE)
  
  ## Temporal EP
  ntmpp <- plot.temporal.p.value(mcmc, t = 12, label = c(41, 48, 100, 182),
                                 file = (if (print) "dlm_pred_dec_clusters.eps" else ""))
}
